"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const qs = require("qs");
const methods_1 = require("./methods");
const request_1 = require("./request");
const utils_1 = require("./utils");
const INSTANCES = {};
function resolve(key) {
    if (!INSTANCES[key]) {
        INSTANCES[key] = create();
    }
    return INSTANCES[key];
}
function create(options) {
    const client = {
        options: options || {}
    };
    client._count = 0;
    client._progress = 1;
    client._tasks = [];
    client.create = create;
    client.resolve = resolve;
    client.setOptions = (opts) => {
        Object.assign(client.options, opts);
    };
    client.on = (event, hook) => {
        let name = `on${event[0].toUpperCase()}${event.substr(1)}`;
        if (!client.options[name]) {
            client.options[name] = hook;
            return client;
        }
        if (Array.isArray(client.options[name])) {
            client.options[name].push(hook);
        }
        else {
            client.options[name] = [client.options[name], hook];
        }
        return client;
    };
    client.off = (event, hook) => {
        let name = `on${event[0].toUpperCase()}${event.substr(1)}`;
        if (!client.options[name]) {
            return client;
        }
        if (Array.isArray(client.options[name])) {
            client.options[name] = client.options[name].filter((f) => f !== hook);
            if (!client.options[name].length)
                client.options[name] = null;
        }
        else if (client.options[name] === hook) {
            client.options[name] = null;
        }
        return client;
    };
    client.getFormDataClass = function getFormDataClass() {
        let FormData = client.options.FormData;
        if (!FormData && typeof window === 'object' && window.FormData) {
            FormData = window.FormData;
        }
        else if (!FormData && typeof global === 'object' && global.FormData) {
            FormData = global.FormData;
        }
        return FormData;
    };
    client.createBody = function (body) {
        if (!body || typeof body !== 'object' || (0, utils_1.isUint8Array)(body) || (0, utils_1.isReadableStream)(body))
            return body;
        if (typeof ArrayBuffer === 'function' && body instanceof ArrayBuffer && typeof Uint8Array === 'function') {
            return new Uint8Array(body);
        }
        let FormData = client.getFormDataClass();
        if (!FormData || body instanceof FormData)
            return body;
        let form = null;
        function addField(path, value) {
            if (value === undefined)
                return;
            if (!(0, utils_1.isReadableStream)(value) && !(0, utils_1.isFile)(value) && !(0, utils_1.isUint8Array)(value)) {
                if (typeof value === 'boolean' || typeof value === 'number' || value === null || value instanceof Date) {
                    value = String(value);
                }
                else if (typeof value === 'object') {
                    for (let key in value) {
                        addField(`${path}[${key}]`, value[key]);
                    }
                    return;
                }
            }
            form.append(path, value);
        }
        for (let key in body) {
            let value = body[key];
            if ((0, utils_1.isReadableStream)(value) || (0, utils_1.isFile)(value) || (0, utils_1.isUint8Array)(value)) {
                form = new FormData();
                Object.keys(body).forEach((path) => addField(path, body[path]));
                break;
            }
        }
        return form || body;
    };
    client.request = function request(path, init, reducer) {
        init = Object.assign({}, init);
        let queryParams = Object.assign({}, init.query);
        delete init.query;
        let queryString = qs.stringify(queryParams, client.options.qsOptions);
        if (queryString) {
            if (path.indexOf('?') < 0) {
                queryString = `?${queryString}`;
            }
            else if (path[path.length - 1] !== '&' && path[path.length - 1] !== '?') {
                queryString = `&${queryString}`;
            }
            path += queryString;
        }
        let apiRoot = client.options.apiRoot;
        if (apiRoot) {
            if (apiRoot[apiRoot.length - 1] === '/' && path[0] === '/') {
                path = path.substring(1);
            }
            else if (apiRoot[apiRoot.length - 1] !== '/' && path[0] !== '/') {
                path = `/${path}`;
            }
            path = apiRoot + path;
        }
        let fetch = client.options.fetch;
        if (!fetch) {
            if (typeof window !== 'undefined') {
                fetch = window.fetch.bind(window);
            }
            else if (typeof global !== 'undefined') {
                fetch = global.fetch;
            }
        }
        let req = new request_1.default(client, fetch, path, init, reducer);
        client._count += 1;
        if (client.options.onProgress) {
            client._tasks.push(req);
        }
        client._updateProgress();
        return req;
    };
    client._updateProgress = function () {
        if (!client.options.onProgress)
            return;
        let now = Date.now();
        client._tasks = client._tasks.filter((t) => !t._endAt || now - t._endAt < 1000);
        if (client._tasks.find((t) => t._endAt)) {
            setTimeout(() => {
                client._updateProgress();
            }, 1000);
        }
        if (client._updateProgressTimer)
            return;
        client._updateProgressTimer = setTimeout(() => {
            client._updateProgressTimer = 0;
            if (!client.options.onProgress)
                return;
            let process = 1;
            let total = client._tasks.length * 3;
            if (total) {
                let steps = 0;
                client._tasks.forEach((r) => {
                    steps += r._endAt ? 3 : r._steps;
                });
                process = steps / total;
            }
            if (process !== this._progress) {
                this._progress = process;
                if (Array.isArray(client.options.onProgress)) {
                    client.options.onProgress.forEach((fn) => fn(process));
                }
                else {
                    client.options.onProgress(process);
                }
            }
        }, 5);
    };
    methods_1.default.forEach((method) => {
        client[method] = function (path, init) {
            init = init || {};
            init.method = method.toUpperCase();
            return client.request(path, init);
        };
    });
    return client;
}
exports.default = resolve('default');
