"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Debugger = require("debug");
const qs = require("qs");
const json_stream_1 = require("./json-stream");
const utils_1 = require("./utils");
const debug = Debugger('akita:request');
function execHooks(request, hooks) {
    let promise;
    if (Array.isArray(hooks)) {
        hooks.forEach((h) => {
            if (promise) {
                promise = promise.then(() => {
                    let p = h(request);
                    return p ? p : Promise.resolve();
                });
            }
            else {
                promise = h(request);
            }
        });
    }
    else {
        promise = hooks(request);
    }
    if (promise) {
        return promise.then(() => Promise.resolve());
    }
}
class Request {
    constructor(client, fetch, url, init, reducer) {
        this.client = client;
        this._reducer = reducer;
        this._fetch = fetch;
        this.url = url;
        this.init = init || {};
        this._steps = 0;
        let promise;
        if (init.body && client.options.onEncode) {
            debug('exec onEncode');
            promise = execHooks(this, client.options.onEncode);
        }
        if (client.options.onRequest) {
            debug('exec onRequest');
            if (promise) {
                promise = promise.then(() => execHooks(this, client.options.onRequest));
            }
            else {
                promise = execHooks(this, client.options.onRequest);
            }
        }
        if (promise) {
            promise = promise.then(() => this._send());
        }
        else {
            promise = this._send();
        }
        this._responsePromise = promise;
    }
    _addStep() {
        this._steps += 1;
        if (this._steps >= 3) {
            this._endAt = Date.now();
        }
        this.client._updateProgress();
    }
    _end() {
        this._endAt = Date.now();
        this.client._updateProgress();
    }
    _send() {
        let { client, url, init } = this;
        let headers = init.headers || {};
        if (client.options.init) {
            init = Object.assign({}, client.options.init, init);
            if (client.options.init.headers) {
                headers = Object.assign({}, client.options.init.headers, headers);
            }
        }
        init.headers = headers;
        this.init = init;
        if (debug.enabled) {
            debug(init.method, url, JSON.stringify(init));
        }
        if (init.body && typeof init.body === 'object') {
            init.body = client.createBody(init.body);
            let FormData = client.getFormDataClass();
            if (!(0, utils_1.isUint8Array)(init.body) && !(FormData && init.body instanceof FormData) && !(0, utils_1.isReadableStream)(init.body)) {
                if (!init.headers) {
                    init.headers = {};
                }
                if (!init.headers['Content-Type'] || init.headers['Content-Type'].indexOf('json') > -1) {
                    if (!init.headers['Content-Type']) {
                        init.headers['Content-Type'] = 'application/json';
                    }
                    init.body = JSON.stringify(init.body);
                }
                else if (init.headers['Content-Type'] === 'application/x-www-form-urlencoded') {
                    init.body = qs.stringify(init.body);
                }
                else {
                    throw new Error(`Akita Error: Unsupported Content-Type ${init.headers['Content-Type']}`);
                }
            }
        }
        this._addStep();
        if (debug.enabled) {
            debug('fetch', url, JSON.stringify(init));
        }
        let promise = this._fetch(url, init).then((res) => {
            this._addStep();
            debug('response:', this.url, res.status, res.statusText);
            this.res = res;
            if (client.options.onResponse) {
                debug('exec onResponse');
                let p = execHooks(this, client.options.onResponse);
                if (p) {
                    return p.then(() => {
                        return Promise.resolve(res);
                    });
                }
            }
            return Promise.resolve(res);
        }, (error) => {
            debug('fetch error:', error.message);
            this._end();
            return Promise.reject(error);
        });
        return promise;
    }
    response() {
        if (!this._endAt)
            this._end();
        return this._responsePromise;
    }
    stream() {
        if (this.res)
            return Promise.resolve(this.res.body);
        return this._responsePromise.then((res) => {
            this._end();
            return res.body;
        });
    }
    jsonStream() {
        if (!this._jsPromise) {
            this._jsPromise = this.stream().then((stream) => new json_stream_1.default(stream, this._reducer));
        }
        return this._jsPromise;
    }
    ok() {
        if (this.res)
            return Promise.resolve(this.res.ok);
        return this._responsePromise.then((res) => res.ok);
    }
    status() {
        if (this.res)
            return Promise.resolve(this.res.status);
        return this._responsePromise.then((res) => res.status);
    }
    statusText() {
        if (this.res)
            return Promise.resolve(this.res.statusText);
        return this._responsePromise.then((res) => res.statusText);
    }
    size() {
        if (this.res)
            return Promise.resolve(this.res.size || parseInt(this.res.headers.get('Content-Length')) || 0);
        return this._responsePromise.then((res) => res.size || parseInt(res.headers.get('Content-Length')) || 0);
    }
    headers() {
        if (this.res)
            return Promise.resolve(this.res.headers);
        return this._responsePromise.then((res) => res.headers);
    }
    buffer() {
        if (!this._bufferPromise) {
            this._bufferPromise = this._responsePromise.then((res) => {
                let fn = 'buffer';
                if (res.arrayBuffer) {
                    fn = 'arrayBuffer';
                }
                return res[fn]().then((buf) => {
                    if (!(0, utils_1.isUint8Array)(buf)) {
                        if (typeof Buffer === 'function') {
                            return Buffer.from(buf);
                        }
                        return buf;
                    }
                    this._addStep();
                    return buf;
                }, (error) => {
                    debug('get buffer error:', error.message);
                    this._end();
                    return Promise.reject(error);
                });
            });
        }
        return this._bufferPromise;
    }
    blob() {
        if (!this._blobPromise) {
            this._blobPromise = this._responsePromise.then((res) => res.blob().then((blob) => {
                this._addStep();
                return blob;
            }, (error) => {
                debug('get blob error:', error.message);
                this._end();
                return Promise.reject(error);
            }));
        }
        return this._blobPromise;
    }
    text() {
        if (!this._textPromise) {
            if (this.raw) {
                this._textPromise = Promise.resolve(this.raw);
            }
            else {
                this._textPromise = this._responsePromise.then((res) => res.text().then((text) => {
                    debug('response text:', text);
                    this.raw = text;
                    this._addStep();
                    return text;
                }, (error) => {
                    debug('get text error:', error.message);
                    this._end();
                    return Promise.reject(error);
                }));
            }
        }
        return this._textPromise;
    }
    _decode() {
        if (this.value !== undefined)
            return this.value;
        debug('_decode');
        let value;
        try {
            value = JSON.parse(this.raw);
        }
        catch (error) {
            return Promise.reject(new Error(`invalid json response body at ${this.url} ${error.message}`));
        }
        if ((value === null || value === void 0 ? void 0 : value.error) && ['0', 'null', 'none'].indexOf(value.error) < 0) {
            let error = new Error(value.error);
            Object.keys(value).forEach((key) => {
                if (['error', 'message', 'stack'].indexOf(key) === -1) {
                    error[key] = value[key];
                }
            });
            return Promise.reject(error);
        }
        this.value = value;
        return value;
    }
    data() {
        if (!this._dataPromise) {
            if (this.value !== undefined) {
                this._dataPromise = Promise.resolve(this.value);
            }
            else {
                this._dataPromise = this._responsePromise.then((res) => {
                    if (res.status === 204) {
                        this._addStep();
                        return Promise.resolve(null);
                    }
                    return res.text().then((text) => {
                        this._end();
                        debug('response text:', text);
                        this.raw = text;
                        const client = this.client;
                        if (client.options.onDecode) {
                            debug('exec onDecode');
                            let promise = execHooks(this, client.options.onDecode);
                            if (promise) {
                                return promise.then(() => Promise.resolve(this._decode()));
                            }
                        }
                        return this._decode();
                    }, (error) => {
                        debug('get text raw error:', error.message);
                        this._end();
                        return Promise.reject(error);
                    });
                });
            }
        }
        return this._dataPromise;
    }
    then(onSuccess, onFail) {
        if (this._reducer) {
            return this.data().then((json) => onSuccess(this._reducer(json)), onFail);
        }
        return this.data().then(onSuccess, onFail);
    }
    catch(onFail) {
        return this.data().catch(onFail);
    }
    finally(fn) {
        return this.data().finally(fn);
    }
}
exports.default = Request;
